import { Collection, InteractionResponse, Message, SnowflakeUtil } from "discord.js";

import { openai } from "./index";
import { formatRequestOrResponse } from "./toOpenAIMessages";
import { config } from "./index";

export default class Moderation {
  /** Represents cache of messages that have been checked aganist OpenAI moderation API. */
  private static cache = new Collection<string, boolean>();
  private static worker: NodeJS.Timeout | null = null;

  /**
   * Checks the message in the OpenAI Moderation API
   * @param message The message to check
   * @returns if the message fails the moderation (true) or not (false)
   */
  public static async checkMessage(message: Message | InteractionResponse): Promise<boolean> {
    if (this.cache.has(message.id)) {
      return this.cache.get(message.id) as boolean;
    }

    if (message instanceof Message) {
      const warningReaction = message.reactions.resolve("⚠");
      // if bot reacted to that message, we already know that it returned true for moderation API
      if (warningReaction && warningReaction.me) {
        this.cache.set(message.id, true);
        return true;
      }
    }
    
    const answer = await openai.moderations.create({
      input: await formatRequestOrResponse(message),
    });

    const flagged = answer.results[0].flagged;
    this.cache.set(message.id, flagged);
    // FIXME: These next 7 lines does not belong there and should be refactored out.
    if (flagged) if (message instanceof Message) {
      message.react("⚠").catch(() => { /* GRACEFAIL: We don't inform the enduser then */ });
    }
    else {
      const channelMessage = await message.fetch();
      channelMessage.react("⚠").catch(() => { /* GRACEFAIL: We don't inform the enduser then */ });
    }

    return flagged;
  }

  /**
   * Checks the message in the OpenAI ModerationAPI
   * @param message The message to check
   */
  public static checkMessageNoReturn(message: Message | InteractionResponse) {
    this.checkMessage(message).catch(e => {
      console.error(`Error occured while handling moderation request (${(e as object).constructor.name}):`);
      console.error(e);
    });
  }

  public static removeExpiredCacheEntries() {
    const now = Date.now();
    for (const i of this.cache.keys()) {
      if (now - SnowflakeUtil.timestampFrom(i) >= config.readLimits.time * 2) {
        this.cache.delete(i);
      }
    }
  }
  public static startCacheClearWorker() {
    if (this.worker) return;
    this.worker = setInterval(() => {
      this.removeExpiredCacheEntries();
    },
    1000 * 60 * 15 /* 15 minutes */,
    );
  }

  public static stopCacheClearWorker() {
    if (!this.worker) return;
    clearInterval(this.worker);
    this.worker = null;
  }
}

Moderation.startCacheClearWorker();
